/*
 * Exercise: Implementation of a vector clock
 *
 * Given scenario: several servers can process tasks specified by a client. The
 * servers work on the task for a while and then forward it to a randomly
 * selected other server. Thus the tasks are never finished!
 *
 * Realize a vector clock and extend the server code and the method log() so
 * that the current vector time is also output for each event. The client should
 * *not* be included in the vector time.
 */

import java.rmi.*;
import java.util.Random;
import java.util.LinkedList;
import java.util.List;


class VectorClock
{
	int[] time;		// Number of events in each process
	int   myNo;		// Own process number
	int   nProcs;	// Total number of processes

	/**
	 * Constructor. Arguments: own process number, total number of processes.
	 */
	public VectorClock(int myNo, int nProcs)
	{
		this.myNo = myNo;
		this.nProcs = nProcs;
		time = new int[nProcs];
		// Initialize the vector time (strictly speaking, this would not be necessary with Java)
		for(int i=0; i<nProcs; i++)
			time[i] = 0;
	}

	/**
	 * Increment the clock, e.g., when an event occurs.
	 */
	synchronized public void tick()
	{
		// ... Add your solution here ... 
	}

	/**
	 * Update the vector time with the time stamp received in a message.
	 */
	synchronized public void update(int[] receivedTime)
	{
		// ... Add your solution here ... 
	}

	/**
	 * Return the current vector time as a string.
	 */
	public String toString()
	{
		String result = "" + time[0];
		for (int i=1; i<nProcs; i++)
			result += "," + time[i];
		return result;	
	}

	/**
	 * Append the current vector time to the string 'message'.
	 */
	public String addToMessage(String message)
	{
		return message + "@" + this;
	}

	/**
	 * Remove the appended vector time from the String 'message' and
	 * return it in array 'time'.
	 */
	public String extractFromMessage(String message, int[] time)
	{
		String[] parts = message.split("@");
		if (parts.length > 1) {
			String[] elems = parts[1].split(",");
			for (int i=0; i<elems.length; i++)
				time[i] = Integer.parseInt(elems[i]);
		}
		return parts[0];
	}

}

/**
 * Implementation of the server process.
 */
public class Server extends BaseServer
{
	List<String> tasks;           // Set of tasks processed by this server
	Random rnd;                   // Random number generator
	/**
	 * Initialize the server process with number 'me'; the total number of processes is 'num'.
	 */
	public Server(int me, int num) throws RemoteException
	{
		super(me, num);
		rnd = new Random();
		tasks = new LinkedList<String>();
	}

	/**
	 * This method simulates the execution of a task.
	 * In this example, the server only sleeps for a while and then passes the
	 * task to a randomly selected server with a 50% probability. This has the
	 * advantage that the total number of tasks in the system is known and does
	 * not change, so when you implement a snapshot algorithm you can easily verify
	 * the result.
	 */
	void doSomeWork()
	{
		try {
			Thread.sleep(500+rnd.nextInt(500));
		}
		catch (InterruptedException e) {
		}

		// If at least one task is available
		if (!tasks.isEmpty()) {
			// Select a task
			String task = tasks.get(0);
			// Random number between 0 and 2*numServers-1
			int n = rnd.nextInt(2*numServers);
			if (n < numServers) {
			    // With a 50% probability, we get n < numServers. In this case, send a
				// task to the selected server and remove it from the set 'tasks'.
				log("Moving " + task + " to server " + n);
				send(n, task);
				tasks.remove(0);
			}
		}
	}

	/**
	 * Main loop of server process.
	 */
	void execute()
	{
		log("Starting");
		while(true) {
			// Message available?
			Message message = receive(false);
			if (message == null) {
				// No: do some work ...
				log("No message --> working");
				doSomeWork();
			}
			else {
				String msg = message.getContents();
				// Add new task to set 'tasks'.
				log("New task: " + msg);
				tasks.add(msg);
			}
		}
	}

	/**
	 * Print a log message.
	 */
	void log(String s)
	{
		System.out.println("Server " + serverNo + ": " + s);
	}

	/**
	 * Main program: create server object, register it with the RMI registry and
	 * execute the main loop.
	 * Command line arguments: <Number of this server> <Total number of servers>
	 */
	public static void main(String args[])
	{
		try {
			int me = Integer.parseInt(args[0]);
			int num = Integer.parseInt(args[1]);
			Server obj = new Server(me, num);
			obj.execute();
		}
		catch (Exception e) {
			System.out.println("Error: " + e.getMessage());
			e.printStackTrace();
		}
	}
}
